@@ -30,6 +30,10 @@
 #include "gdkinternals.h"
 #include "gdkintl.h"
 
+#include "gdkresources.h"
+
+#include "gdk-private.h"
+
 #ifndef HAVE_XCONVERTCASE
 #include "gdkkeysyms.h"
 #endif
@@ -128,27 +132,40 @@ static int gdk_initialized = 0;                     /* 1 if the library is initi
                                                      */
 
 static gchar  *gdk_progclass = NULL;
+static gboolean gdk_progclass_overridden;
 
 static GMutex gdk_threads_mutex;
 
 static GCallback gdk_threads_lock = NULL;
 static GCallback gdk_threads_unlock = NULL;
 
+static const GDebugKey gdk_gl_keys[] = {
+  { "disable",               GDK_GL_DISABLE },
+  { "always",                GDK_GL_ALWAYS },
+  { "software-draw",         GDK_GL_SOFTWARE_DRAW_GL | GDK_GL_SOFTWARE_DRAW_SURFACE} ,
+  { "software-draw-gl",      GDK_GL_SOFTWARE_DRAW_GL },
+  { "software-draw-surface", GDK_GL_SOFTWARE_DRAW_SURFACE },
+  { "texture-rectangle",     GDK_GL_TEXTURE_RECTANGLE },
+  { "legacy",                GDK_GL_LEGACY },
+  { "gles",                  GDK_GL_GLES },
+};
+
 #ifdef G_ENABLE_DEBUG
 static const GDebugKey gdk_debug_keys[] = {
-  {"events",        GDK_DEBUG_EVENTS},
-  {"misc",          GDK_DEBUG_MISC},
-  {"dnd",           GDK_DEBUG_DND},
-  {"xim",           GDK_DEBUG_XIM},
-  {"nograbs",       GDK_DEBUG_NOGRABS},
-  {"input",         GDK_DEBUG_INPUT},
-  {"cursor",        GDK_DEBUG_CURSOR},
-  {"multihead",     GDK_DEBUG_MULTIHEAD},
-  {"xinerama",      GDK_DEBUG_XINERAMA},
-  {"draw",          GDK_DEBUG_DRAW},
-  {"eventloop",     GDK_DEBUG_EVENTLOOP},
-  {"frames",        GDK_DEBUG_FRAMES},
-  {"settings",      GDK_DEBUG_SETTINGS}
+  { "events",        GDK_DEBUG_EVENTS },
+  { "misc",          GDK_DEBUG_MISC },
+  { "dnd",           GDK_DEBUG_DND },
+  { "xim",           GDK_DEBUG_XIM },
+  { "nograbs",       GDK_DEBUG_NOGRABS },
+  { "input",         GDK_DEBUG_INPUT },
+  { "cursor",        GDK_DEBUG_CURSOR },
+  { "multihead",     GDK_DEBUG_MULTIHEAD },
+  { "xinerama",      GDK_DEBUG_XINERAMA },
+  { "draw",          GDK_DEBUG_DRAW },
+  { "eventloop",     GDK_DEBUG_EVENTLOOP },
+  { "frames",        GDK_DEBUG_FRAMES },
+  { "settings",      GDK_DEBUG_SETTINGS },
+  { "opengl",        GDK_DEBUG_OPENGL }
 };
 
 static gboolean
@@ -196,6 +213,7 @@ static gboolean
 gdk_arg_class_cb (const char *key, const char *value, gpointer user_data, GError **error)
 {
   gdk_set_program_class (value);
+  gdk_progclass_overridden = TRUE;
 
   return TRUE;
 }
@@ -215,9 +233,11 @@ static const GOptionEntry gdk_args[] = {
   { "name",         0, 0,                     G_OPTION_ARG_CALLBACK, gdk_arg_name_cb,
     /* Description of --name=NAME in --help output */          N_("Program name as used by the window manager"),
     /* Placeholder in --name=NAME in --help output */          N_("NAME") },
+#ifndef G_OS_WIN32
   { "display",      0, G_OPTION_FLAG_IN_MAIN, G_OPTION_ARG_STRING,   &_gdk_display_name,
     /* Description of --display=DISPLAY in --help output */    N_("X display to use"),
     /* Placeholder in --display=DISPLAY in --help output */    N_("DISPLAY") },
+#endif
 #ifdef G_ENABLE_DEBUG
   { "gdk-debug",    0, 0, G_OPTION_ARG_CALLBACK, gdk_arg_debug_cb,  
     /* Description of --gdk-debug=FLAGS in --help output */    N_("GDK debugging flags to set"),
@@ -229,26 +249,54 @@ static const GOptionEntry gdk_args[] = {
   { NULL }
 };
 
+void
+gdk_add_option_entries (GOptionGroup *group)
+{
+  g_option_group_add_entries (group, gdk_args);
+}
+
 /**
  * gdk_add_option_entries_libgtk_only:
  * @group: An option group.
  *
  * Appends gdk option entries to the passed in option group. This is
  * not public API and must not be used by applications.
+ *
+ * Deprecated: 3.16: This symbol was never meant to be used outside
+ *   of GTK+
  */
 void
 gdk_add_option_entries_libgtk_only (GOptionGroup *group)
 {
-  g_option_group_add_entries (group, gdk_args);
+  gdk_add_option_entries (group);
+}
+
+static gpointer
+register_resources (gpointer dummy G_GNUC_UNUSED)
+{
+  _gdk_register_resource ();
+
+  return NULL;
+}
+
+static void
+gdk_ensure_resources (void)
+{
+  static GOnce register_resources_once = G_ONCE_INIT;
+
+  g_once (&register_resources_once, register_resources, NULL);
 }
 
 void
-gdk_pre_parse_libgtk_only (void)
+gdk_pre_parse (void)
 {
   const char *rendering_mode;
+  const gchar *gl_string;
 
   gdk_initialized = TRUE;
 
+  gdk_ensure_resources ();
+
   /* We set the fallback program class here, rather than lazily in
    * gdk_get_program_class, since we don't want -name to override it.
    */
@@ -266,6 +314,12 @@ gdk_pre_parse_libgtk_only (void)
   }
 #endif  /* G_ENABLE_DEBUG */
 
+  gl_string = getenv("GDK_GL");
+  if (gl_string != NULL)
+    _gdk_gl_flags = g_parse_debug_string (gl_string,
+                                          (GDebugKey *) gdk_gl_keys,
+                                          G_N_ELEMENTS (gdk_gl_keys));
+
   if (getenv ("GDK_NATIVE_WINDOWS"))
     {
       g_warning ("The GDK_NATIVE_WINDOWS environment variable is not supported in GTK3.\n"
@@ -285,6 +339,20 @@ gdk_pre_parse_libgtk_only (void)
     }
 }
 
+/**
+ * gdk_pre_parse_libgtk_only:
+ *
+ * Prepare for parsing command line arguments for GDK. This is not
+ * public API and should not be used in application code.
+ *
+ * Deprecated: 3.16: This symbol was never meant to be used outside
+ *   of GTK+
+ */
+void
+gdk_pre_parse_libgtk_only (void)
+{
+  gdk_pre_parse ();
+}
   
 /**
  * gdk_parse_args:
@@ -313,7 +381,7 @@ gdk_parse_args (int    *argc,
   if (gdk_initialized)
     return;
 
-  gdk_pre_parse_libgtk_only ();
+  gdk_pre_parse ();
 
   option_context = g_option_context_new (NULL);
   g_option_context_set_ignore_unknown_options (option_context, TRUE);
@@ -372,8 +440,8 @@ gdk_get_display_arg_name (void)
    return _gdk_display_arg_name;
 }
 
-/**
- * gdk_display_open_default_libgtk_only:
+/*< private >
+ * gdk_display_open_default:
  *
  * Opens the default display specified by command line arguments or
  * environment variables, sets it as the default display, and returns
@@ -383,9 +451,9 @@ gdk_get_display_arg_name (void)
  *
  * Returns: (nullable) (transfer none): the default display, if it
  *   could be opened, otherwise %NULL.
- **/
+ */
 GdkDisplay *
-gdk_display_open_default_libgtk_only (void)
+gdk_display_open_default (void)
 {
   GdkDisplay *display;
 
@@ -400,6 +468,53 @@ gdk_display_open_default_libgtk_only (void)
   return display;
 }
 
+gboolean
+gdk_running_in_sandbox (void)
+{
+  return g_file_test ("/.flatpak-info", G_FILE_TEST_EXISTS);
+}
+
+gboolean
+gdk_should_use_portal (void)
+{
+  static const char *use_portal = NULL;
+
+  if (G_UNLIKELY (use_portal == NULL))
+    {
+      if (gdk_running_in_sandbox ())
+        use_portal = "1";
+      else
+        {
+          use_portal = g_getenv ("GTK_USE_PORTAL");
+          if (!use_portal)
+            use_portal = "";
+        }
+    }
+
+  return use_portal[0] == '1';
+}
+
+/**
+ * gdk_display_open_default_libgtk_only:
+ *
+ * Opens the default display specified by command line arguments or
+ * environment variables, sets it as the default display, and returns
+ * it. gdk_parse_args() must have been called first. If the default
+ * display has previously been set, simply returns that. An internal
+ * function that should not be used by applications.
+ *
+ * Returns: (nullable) (transfer none): the default display, if it
+ *   could be opened, otherwise %NULL.
+ *
+ * Deprecated: 3.16: This symbol was never meant to be used outside
+ *   of GTK+
+ */
+GdkDisplay *
+gdk_display_open_default_libgtk_only (void)
+{
+  return gdk_display_open_default ();
+}
+
 /**
  * gdk_init_check:
  * @argc: (inout): the number of command line arguments.
@@ -422,7 +537,7 @@ gdk_init_check (int    *argc,
 {
   gdk_parse_args (argc, argv);
 
-  return gdk_display_open_default_libgtk_only () != NULL;
+  return gdk_display_open_default () != NULL;
 }
 
 
@@ -512,7 +627,6 @@ gdk_init (int *argc, char ***argv)
  * management for you.
  */
 
-
 /**
  * gdk_threads_enter:
  *
@@ -962,10 +1076,16 @@ gdk_get_program_class (void)
  * Sets the program class. The X11 backend uses the program class to set
  * the class name part of the `WM_CLASS` property on
  * toplevel windows; see the ICCCM.
+ *
+ * The program class can still be overridden with the --class command
+ * line option.
  */
 void
 gdk_set_program_class (const char *program_class)
 {
+  if (gdk_progclass_overridden)
+    return;
+
   g_free (gdk_progclass);
 
   gdk_progclass = g_strdup (program_class);
